# vim:fileencoding=utf-8
import logging

from django.http import HttpResponse
from django.views.generic import View

from . import tsc
from .route import Router

logger = logging.getLogger(__name__)


class RouterJSView(View):
    def get(self, *args, **kwargs):
        logger.debug("Generates router.js")
        ts_src = Router().export()
        js_src = tsc.transpile(ts_src)
        return HttpResponse(js_src, content_type='text/javascript')



# vim:fileencoding=utf-8
from django.conf.urls import url
from django.views.decorators.cache import cache_page

from . import views

urlpatterns = [
    url(r'^router.js$', cache_page(300)(views.RouterJSView.as_view()), name='django_ts_router_js'),
]



# vim:fileencoding=utf-8
import logging

from django.core.management import BaseCommand

from django_ts_router import tsc
from django_ts_router.route import Router

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Generates a TypeScript code for URL reverse function."

    def add_arguments(self, parser):
        parser.add_argument('--out',
                            action='store',
                            dest='out',
                            default=None,
                            help='Output a generated TypeScript code to the given file.')

        parser.add_argument('--module',
                            action='store',
                            dest='module',
                            default=None,
                            help='A module name to which the reverse() function belongs.')

        parser.add_argument('--javascript', '-j',
                            action='store_true',
                            dest='js',
                            default=False,
                            help='Generate a JavaScript code. "--out" option must be specified.')

    def handle(self, *args, **options):
        ts_src = Router().export(module=options['module'])

        if options['out']:
            if options['js']:
                tsc.run(ts_src, options['out'])
            else:
                with open(options['out'], 'w') as fp:
                    fp.write(ts_src)
        else:
            if options['js']:
                self.stdout.write(tsc.transpile(ts_src))
            else:
                self.stdout.write(ts_src)



# vim:fileencoding=utf-8
import json
import logging
import re

from django.conf import settings
from django.contrib.admindocs.views import simplify_regex
from django.template.loader import render_to_string
from django.utils.functional import cached_property

logger = logging.getLogger(__name__)

TAB_SPACE_LENGTH = 4
TAB_SPACES = ' ' * TAB_SPACE_LENGTH


class RoutePattern(object):
    def __init__(self, name, pattern, pattern_prefix=''):
        """
        :param str name:
        :param RegexURLPattern pattern:
        :param str base:
        """
        self._name = name
        self._base = pattern_prefix
        self._pattern = pattern

    @cached_property
    def simplified_pattern(self):
        """
        :rtype: str
        """
        pattern = self._base + self._pattern.regex.pattern
        return simplify_regex(pattern)

    @cached_property
    def pattern_args(self):
        """
        :rtype: list[str]
        """
        return re.findall(r'<(.+?)>', self.simplified_pattern)

    def __repr__(self):
        return "<RoutePattern {name} {pattern}".format(name=self._name, pattern=self.simplified_pattern)

    def to_dict(self):
        """
        :rtype: dict
        """
        return {
            self._name: {
                'pattern': self.simplified_pattern,
                'arguments': self.pattern_args,
            }
        }

    def export_name_to_ts(self):
        return 'export var {constant}:string = "{name}";'.format(
            constant=self._get_constant(),
            name=self._name
        )

    def _get_constant(self):
        return self._name.upper().replace(':', '_').replace('-', '_')

    def generate_function(self):
        """

        :rtype: str
        """
        function_name = self.get_function_name()
        arguments_with_type = [arg + ':string' for arg in self.pattern_args]
        arguments = ', '.join(arguments_with_type)

        body = 'var args:{[arg: string]: string} = {};\n'
        for arg in self.pattern_args:
            body += 'args["%s"] = %s;\n' % (arg, arg)
        body += 'return reverse(%s, args);' % self._get_constant()
        return 'export function %s(%s):string {\n%s\n}' % (function_name, arguments, body)

    def get_function_name(self):
        """Generates a function name 'getXxxxURL'.

        Converts snake case (snake_case) to lower camel case (lowerCamelCase).
        http://stackoverflow.com/a/19053800/3073708

        :type: str
        """
        name = self._name.replace(':', '_').replace('-', '_')
        components = name.split('_')
        return 'get%sURL' % ''.join(x.title() for x in components)


class Router(object):
    DEFAULT_MODULE = 'django.tsrouter'

    def __init__(self):
        urlconf = __import__(settings.ROOT_URLCONF, fromlist=['urlpatterns'])
        self._route_patterns = self.parse_urlpatterns(urlconf.urlpatterns)

        router_settings = getattr(settings, 'TS_ROUTER', None)
        if not router_settings:
            raise RuntimeError('TS_ROUTER setting is not found in your settings file.')
        self.module = router_settings.get('MODULE', self.DEFAULT_MODULE)
        self.names = router_settings.get('NAMES')
        if not self.names:
            raise RuntimeError('One or more NAMES should be specified.')

    def find_route_pattern_by_name(self, name):
        """
        :rtype: RoutePattern
        """
        return self._route_patterns[name]

    def find_names_by_namespace(self, namespace):
        names = list(self._route_patterns)
        namespace += ':'
        matched = filter(lambda n: n.startswith(namespace), names)
        return list(matched)

    @classmethod
    def parse_urlpatterns(cls, urlpatterns, pattern_prefix='', namespace=None, route_patterns=None):
        """

        Reference: django.contrib.admindocs.views.extract_views_from_urlpatterns

        :rtype: dict[str, RoutePattern]
        """
        if not route_patterns:
            route_patterns = {}

        for urlpattern in urlpatterns:
            logger.debug(urlpattern)
            if hasattr(urlpattern, 'url_patterns'):
                logger.debug('hey')
                route_patterns = cls.parse_urlpatterns(urlpattern.url_patterns,
                                                       pattern_prefix=(pattern_prefix + urlpattern.regex.pattern),
                                                       namespace=(urlpattern.namespace or namespace),
                                                       route_patterns=route_patterns)

            elif hasattr(urlpattern, 'callback'):
                if namespace:
                    name = "{namespace}:{name}".format(namespace=namespace, name=urlpattern.name)
                else:
                    name = urlpattern.name

                # Ignore URL pattern without name
                if not name:
                    continue

                route_pattern = RoutePattern(pattern_prefix=pattern_prefix, name=name, pattern=urlpattern)
                route_patterns[name] = route_pattern
        return route_patterns

    def generate_router_typescript(self, names, module):
        """Generates a router TypeScript code for URLs of the given names.

        :param list[str] names: URL names to be output
        :param str module: Module name like "com.example.router"
        :return: TypeScript code
        :rtype: str
        """
        the_names = []
        for name in names:
            if ':*' in name:
                namespace = name.split(':*')[0]
                the_names += self.find_names_by_namespace(namespace)
            else:
                the_names.append(name)

        # Remove duplicated names
        names = set(the_names)
        patterns = [self.find_route_pattern_by_name(name) for name in names]

        var_names = []
        functions = []
        d = {}
        for p in patterns:
            d.update(p.to_dict())
            var_names.append(p.export_name_to_ts())
            functions.append(p.generate_function())
        var_names = ('\n' + TAB_SPACES).join(var_names)
        functions = '\n'.join(functions)
        urlpatterns_json = json.dumps(d, indent=4)
        # Format json string for output as TypeScript source.
        urlpatterns_json = '\n'.join(
            [TAB_SPACES + line if index > 0 else line for index, line in
             enumerate(urlpatterns_json.split('\n'))])

        c = {
            'module': module,
            'names': var_names,
            'functions': functions,
            'url_patterns': urlpatterns_json,
        }
        return render_to_string('django_ts_router/router.ts.tmpl', c)

    def export(self, module=None):
        """Generates a router TypeScript code which belongs to the given module.

        When the 'module' argument is not given, self.module is used.

        :param str module: Module name like "com.example.router"
        :return: TypeScript code
        :rtype: str
        """
        if not module:
            module = self.module
        return self.generate_router_typescript(self.names, module)



# vim:fileencoding=utf-8
import logging
import subprocess
import tempfile
import os
import os.path

from django.conf import settings

logger = logging.getLogger(__name__)

DEFAULT_TSC_PATHS = [
    '/usr/bin/tsc',
    '/usr/local/bin/tsc',
]


def run(ts_src, dest):
    tsc = _find_tsc_command_path()
    if not tsc:
        raise LookupError('tsc command cannot be found. Set tsc path to "TSC" option in TS_ROUTER setting.')

    _, filepath = tempfile.mkstemp(suffix='.ts')
    try:
        with open(filepath, 'w') as f:
            f.write(ts_src)
        subprocess.call([tsc, '--removeComments', '--out', dest, filepath])
    finally:
        os.remove(filepath)


def transpile(ts_src):
    """Transpiles the given TypeScript code to JavaScript code and returns it.

    :return: Transpiled JavaScript code
    :rtype: str
    """
    _, filepath = tempfile.mkstemp(suffix='.js')
    try:
        run(ts_src, filepath)
        with open(filepath, 'r') as f:
            return f.read()
    finally:
        os.remove(filepath)


def _find_tsc_command_path():
    tsc = settings.TS_ROUTER.get('TSC')
    if tsc:
        return tsc

    for path in DEFAULT_TSC_PATHS:
        if os.path.exists(path) and os.access(path, os.X_OK):
            return path



# vim:fileencoding=utf-8

from setuptools import setup, find_packages

setup(name='django-ts-router',
      version='0.7.1',
      description='Generates a TypeScript (or JavaScript) source that includes URL reverse functions for your Django app.',
      url='https://github.com/strippers/django-ts-router',
      license='LGPL',
      author='Tomohiro Otsuka',
      author_email='t.otsuka@gmail.com',
      packages=find_packages(
          exclude=['*tests*', 'testproj']
      ),
      package_data={
          '': ['templates/**/*']
      },
      install_requires=[],
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Web Environment',
          'Framework :: Django :: 1.8',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3.4',
          'Topic :: Internet :: WWW/HTTP',
      ],
      keywords=['django', 'javascript', 'typescript', 'routing', 'url'],
      )
